@@ -9,13 +9,14 @@ __authors__ = 'David Nidever <dnidever@montana.edu>'
 __version__ = '20190723'  # yyyymmdd
 
 import numpy as np
+import copy
 from scipy.spatial import cKDTree
-from . import utils
-from astropy.coordinates import frame_transform_graph
+from scipy.optimize import minimize
+from . import utils,ladfit
+from astropy.coordinates import frame_transform_graph,SkyCoord
 from astropy.coordinates.matrix_utilities import rotation_matrix, matrix_product, matrix_transpose
 import astropy.coordinates as coord
 import astropy.units as u
-from astropy.coordinates import SkyCoord
 
 def rotsph(lon,lat,clon,clat,anode=None,reverse=False,original=False):
     '''
@@ -533,13 +534,17 @@ def xmatch(ra1, dec1, ra2, dec2, dcr=2.0,unique=False,sphere=True):
     ra2/dec2 : array_like
         second dataset, arrays of RA and DEC
         both measured in degrees
+    dcr : float (optional)
+        maximum radius of search, measured in arcsec.
+        This can be an array of the same size as ra1/dec1.
+    unique : boolean, optional
+        Return unique one-to-one matches.  Default is False and
+           allows duplicates.
     sphere : boolean, optional
         The coordinates are spherical in degrees.  Otherwise, the dcr
           is assumed to be in the same units as the input values.
           Default is True.
-    dcr : float (optional)
-        maximum radius of search, measured in arcsec.
-        This can be an array of the same size as ra1/dec1.
+
 
     Returns
     -------
@@ -847,3 +852,133 @@ def mag2gal(mlon,mlat):
     glon,glat = c_glon.l.degree, c_glon.b.degree #subtract off 360 from ms_l
     return glon,glat
 
+def wcsfit(wcs,tab,verbose=False):
+    """
+    Fit the WCS using a catalog of stars with known X/Y and RA/DEC coordinates.
+
+    Parameters
+    ----------
+    wcs : WCS object
+      The WCS object with an estimate of the WCS.
+    tab : table
+      Catalog with x, y, ra, and dec of matched sources.
+    verbose : boolean, optional
+      Print informaton to the screen.  Default is False.
+
+    Returns
+    -------
+    fwcs : WCS object
+      WCS with improved paramters.
+    
+    Example
+    -------
+
+    fwcs = wcsfit(wcs,tab)L
+
+    """
+
+    if len(tab)==0:
+        raise ValueError('Input catalog has no stars')
+    
+    if verbose:
+        print('Fitting WCS with '+str(len(tab))+' stars')
+    
+    coo = SkyCoord(ra=tab['ra'],dec=tab['dec'],unit='deg')
+
+    def newwcs(pars):
+        # pars = [CRVAL1,CRVAL2,CDELT1,CDELT2,PC1_1,PC1_2,PC2_1,PC2_2]
+        # pars = [delta_ra (arcsec), delta_dec (arcsec), cdelt1_scale_change (multiplicative)
+        #  cdelt2_scale_change (multiplicate), rotation (deg)]
+        twcs = copy.deepcopy(wcs)
+        twcs.wcs.crval[0] += pars[0]/3600.0
+        twcs.wcs.crval[1] += pars[1]/3600.0  
+        twcs.wcs.cdelt[0] *= pars[2]
+        twcs.wcs.cdelt[1] *= pars[3]
+        # Rotation matrix
+        # R = [[cos(theta), -sin(theta)], [sin(theta), cos(theta)]]
+        rot = np.array([[np.cos(np.deg2rad(pars[4])),-np.sin(np.deg2rad(pars[4]))],
+                        [np.sin(np.deg2rad(pars[4])),np.cos(np.deg2rad(pars[4]))] ])
+        newpc = np.dot(twcs.wcs.pc,rot)
+        twcs.wcs.pc = newpc
+        return twcs
+    def diffcoords(pars,good=None):
+        twcs = newwcs(pars)
+        try:
+            vcoo = twcs.pixel_to_world(tab['x'],tab['y'])
+            diff = coo.separation(vcoo)
+            if good is not None:
+                meddiff = np.nanmean(diff[good].arcsec)
+            else:
+                meddiff = np.nanmean(diff.arcsec)
+        except:
+            meddiff = 999999.
+        return meddiff
+
+    def diffcoords_jac(pars):
+        delta = [0.2,0.2,0.1,0.1,0.5]
+        jac = np.zeros(5,float)
+        y0 = diffcoords(pars)
+        for i in range(len(pars)):
+            tpars = np.array(pars).copy()
+            tpars[i] += delta[i]
+            y1 = diffcoords(tpars)
+            jac[i] = (y1-y0)/delta[i]
+        return y0,jac
+            
+    
+    # fit delta_ra (arcsec), delta_dec (arcsec), cdelt1_scale_change (multiplicative)
+    #  cdelt2_scale_change (multiplicate), rotation (deg)]
+
+
+    # Get initial guess of delta_ra and delta_dec using the coordinate ra/dec values
+    # and initial WCS coordinates
+    vcoo = wcs.pixel_to_world(tab['x'],tab['y'])
+    vra = vcoo.ra.deg
+    vdec = vcoo.dec.deg
+    dra = np.median(tab['ra']-vra)*3600
+    ddec = np.median(tab['dec']-vdec)*3600
+
+    # Estimate the rotation
+    vx,vy = wcs.world_to_pixel(coo)
+    coefx,absdevx = ladfit.ladfit(tab['y'],tab['x']-vx)
+    coefy,absdevy = ladfit.ladfit(tab['x'],tab['y']-vy)    
+    rot = np.mean([coefx[1],-coefy[1]])
+
+    # Do the fit
+    estimates = [dra,ddec,1.0,1.0,rot]    
+    bounds = len(estimates)*[[-np.inf,np.inf]]
+    res1 = minimize(diffcoords,estimates,bounds=bounds)
+    pars1 = res1.x
+    
+    # Remove outliers and refit
+    twcs = newwcs(pars1)
+    vcoo = twcs.pixel_to_world(tab['x'],tab['y'])
+    diff = coo.separation(vcoo)
+    meddiff = np.nanmedian(diff.arcsec)
+    sigdiff = utils.mad(diff.arcsec)
+    good, = np.where(diff.arcsec < (meddiff+3.5*sigdiff))
+    noutlier = len(tab)-len(good)
+    if noutlier>0:
+        if verbose:
+            print('Rejecting '+str(noutlier)+' outlier(s)')
+        res = minimize(diffcoords,pars1,args=good,bounds=bounds)
+        pars = res.x
+    else:
+        res = res1
+        pars = pars1
+                  
+    fwcs = newwcs(pars)
+    
+    if verbose:
+        print('--- Original WCS ---')
+        print(wcs)
+        print(' ')
+        print('--- Final WCS ---')
+        print(fwcs)
+        print(' ')
+        resid0 = diffcoords([0.0,0.0,1.0,1.0,0.0])
+        resid = diffcoords(pars,good)
+        print('Original mean residuals: {:.3f} arcsec'.format(resid0))
+        print('Final mean residuals   : {:.3f} arcsec'.format(resid))
+        
+    return fwcs
