@@ -12,6 +12,7 @@ import time
 import numpy as np
 import warnings
 from astropy.utils.exceptions import AstropyWarning
+import matplotlib
 import matplotlib.pyplot as plt
 import matplotlib.colors as colors
 from matplotlib.backend_bases import MouseButton
@@ -124,9 +125,11 @@ def zscaling(im,contrast=0.25,nsample=50000):
     
 
 def hist2d(x,y,z=None,statistic=None,xr=None,yr=None,dx=None,dy=None,nx=200,ny=200,
-           zscale=None,log=None,noerase=False,zmin=None,zmax=None,center=True,force=True,cmap=None,
-           xtitle=None,ytitle=None,title=None,colorlabel=None,origin='lower',aspect='auto',
-           bright=None,minhue=0.0,maxhue=0.7,minbright=0.1,maxbright=0.7,saturation=0.9):
+           zscale=None,log=None,norm=None,noerase=False,vmin=None,vmax=None,center=True,
+           xflip=False,yflip=False,force=True,cmap=None,figure=None,
+           figsize=(8,8),xtitle=None,ytitle=None,title=None,colorlabel=None,charsize=12,
+           origin='lower',aspect='auto',interpolation='none',bright=None,minhue=0.0,
+           maxhue=0.7,minbright=0.1,maxbright=0.7,saturation=0.9,noplot=False,save=None):
     """
     Make a 2D histogram of points.
     
@@ -158,17 +161,26 @@ def hist2d(x,y,z=None,statistic=None,xr=None,yr=None,dx=None,dy=None,nx=200,ny=2
        Automatic min and max interval of statistic based on IRAF's zscale.
     log : bool, optional
        Use logarithmic scaling of the statistic.  Default is linear.
+    norm : normalize object, optional
+        The `.Normalize` instance used to scale scalar data to the [0, 1]
+        range before mapping to colors using *cmap*. By default, a linear
+        scaling mapping the lowest value to 0 and the highest to 1 is used.
+        This parameter is ignored for RGB(A) data.
     noerase : bool, optional
        Do not erase or clear the figure window.  Default is to clear the figure.
-    zmin : float, optional
+    vmin : float, optional
        Minimum value to plot on colorbar.  Default is the minimum of the statistic.
-    zmax : float, optional
+    vmax : float, optional
        Maximum value to plot on colorbar.  Default is the maximum of the statistic.    
     force : bool, optional
        Force the xrange and yrange to be exactly as the way they are set (default).
     center : bool, optional
        The x/y values of a bin should correspond to the center.  By default they
          correspond to the bottom-left corner.
+    xflip : bool, optional
+       Flip the X-coordinate axis.  Default is False.
+    yflip : bool, optional
+       Flip the Y-coordinate axis.  Default is False.
     cmap : str, optional
        The matplotlib color map to use.  Default is 'viridis'.
     xtitle : str, optional
@@ -177,12 +189,24 @@ def hist2d(x,y,z=None,statistic=None,xr=None,yr=None,dx=None,dy=None,nx=200,ny=2
        The label for the y-axis.  Default is 'Y'.
     title : str, optional
        The plot title.  Default is statistic+'(Z)'.
+    figure : int, optional
+      The figure window number.  Default is to use the current window.
+    figsize : list, optional
+       Two-element list giving figure size in X and Y.  Default is (8,8).
     colorlabel : str, optional
        The colorbar label.  Default is statistic+'(Z)'.
+    charsize : int, optional
+       Character size.  Default is 12.
     origin : str, optional
        The origin of the plot.  Default is 'lower'.
     aspect : str, optional
        The aspect ratio of the plot.  Default is 'auto'.
+    interpolation : str, optional
+       Type of interpolation in the image.  Default is 'none'.
+    noplot : boolean, optional
+       Do not plot anything.  Just return the data.
+    save : str, optional
+       Save the figure to this file.
 
     Returns
     -------
@@ -197,11 +221,18 @@ def hist2d(x,y,z=None,statistic=None,xr=None,yr=None,dx=None,dy=None,nx=200,ny=2
     """
 
     # Getting the current figure, creating a new one if necessary
-    fig = plt.gcf()
-    
-    if noerase is False:
-        plt.clf()   # clear the plot
+    if figure is None:
+        fig = plt.gcf()
+    else:
+        fig = plt.figure(figure,figsize=figsize)
 
+    # Character size
+    font = {'size': charsize}
+    matplotlib.rc('font', **font)
+
+    if noerase is False:
+        plt.clf()   # clear the plot    
+    
     # Statistic default
     if statistic is None:
         if z is None:
@@ -220,10 +251,10 @@ def hist2d(x,y,z=None,statistic=None,xr=None,yr=None,dx=None,dy=None,nx=200,ny=2
         dy0 = copy.copy(dy)
         
     # Min and Max's
-    xmin = np.min(x)
-    xmax = np.max(x)
-    ymin = np.min(y)
-    ymax = np.max(y)    
+    xmin = np.nanmin(x)
+    xmax = np.nanmax(x)
+    ymin = np.nanmin(y)
+    ymax = np.nanmax(y)    
 
     # Temporary DX and DY if not input
     # X-axis
@@ -386,8 +417,9 @@ def hist2d(x,y,z=None,statistic=None,xr=None,yr=None,dx=None,dy=None,nx=200,ny=2
         yr = [ymin,ymax]
         
     # No z input
-    if z is None:
+    if z is None or statistic=='count':
         im, xedges, yedges = np.histogram2d(x,y,range=[xr,yr],bins=[nx,ny])
+        im = im.T  # np.histogram2d() returns [x,y], while python/matplotlib convention is [y,x]
     # Statistic using z-values
     else:
         if statistic=='avg':
@@ -416,8 +448,8 @@ def hist2d(x,y,z=None,statistic=None,xr=None,yr=None,dx=None,dy=None,nx=200,ny=2
             #ima2 = -ima    ; (blue-green-red)
             # color_convert, hue, bright, im*0.+saturation, r, g, b, /HLS_RGB
             gooda = np.isfinite(ima)
-            if zmin is None: zmin = np.min(ima[gooda])
-            if zmax is None: zmax = np.max(ima[gooda])            
+            if vmin is None: vmin = np.min(ima[gooda])
+            if vmax is None: vmax = np.max(ima[gooda])            
             #ima = -ima
 
             # hue goes: red, orange, yellow, green, light blue, dark blue, purple, red
@@ -426,7 +458,7 @@ def hist2d(x,y,z=None,statistic=None,xr=None,yr=None,dx=None,dy=None,nx=200,ny=2
             # hsv is hue, saturation, value or hue, saturation, brightness
             hueim = np.zeros(ima.shape,float)
             #hueim[gooda] = dln.limit(dln.scale(ima[gooda],[-zmax,-zmin],[minhue,maxhue]),minhue,maxhue)
-            hueim[gooda] = dln.limit(dln.scale(ima[gooda],[zmin,zmax],[minhue,maxhue]),minhue,maxhue)
+            hueim[gooda] = dln.limit(dln.scale(ima[gooda],[vmin,vmax],[minhue,maxhue]),minhue,maxhue)
             # flip hue to get blue to red
             hueim = 1-hueim
             goodt = np.isfinite(imt)
@@ -452,46 +484,59 @@ def hist2d(x,y,z=None,statistic=None,xr=None,yr=None,dx=None,dy=None,nx=200,ny=2
             im = im.T
 
     # Plot the image
-    #fig, ax = plt.subplots()
-    norm = None
-    if log is True and statistic != 'avg':
-        norm = colors.LogNorm(vmin=zmin,vmax=zmax)
-    if zscale is True:
-        zmin,zmax = zscaling(im)
-    if zmin is None:
-        zmin = np.min(im[np.isfinite(im)])
-    if zmax is None:
-        zmax = np.max(im[np.isfinite(im)])        
-    extent = [xr[0], xr[1], yr[0], yr[1]]
-    plt.imshow(im,cmap=cmap,norm=norm,aspect=aspect,vmin=zmin,vmax=zmax,origin=origin,extent=extent)
-    
-    # Axis titles
-    if xtitle is None:
-        xtitle = 'X'
-    plt.xlabel(xtitle)
-    if ytitle is None:
-        ytitle = 'Y'
-    plt.ylabel(ytitle)        
-    if title is None:
-        if statistic is 'count':
-            title = statistic
+    if noplot==False:
+        if log is True and statistic != 'avg' and norm is None:
+            norm = colors.LogNorm(vmin=vmin,vmax=vmax)
+        if zscale is True:
+            vmin,vmax = zscaling(im)
+        if vmin is None:
+            vmin = np.min(im[np.isfinite(im)])
+        if vmax is None:
+            vmax = np.max(im[np.isfinite(im)])        
+        extent = [xr[0], xr[1], yr[0], yr[1]]
+
+        if norm is not None:
+            plt.imshow(im,cmap=cmap,norm=norm,aspect=aspect,origin=origin,
+                       extent=extent,interpolation=interpolation)
         else:
-            title = statistic+'(Z)'
-    plt.title(title)
+            plt.imshow(im,cmap=cmap,norm=norm,aspect=aspect,vmin=vmin,vmax=vmax,origin=origin,
+                       extent=extent,interpolation=interpolation)
+        if xflip: plt.xlim(xr[1],xr[0])
+        if yflip: plt.ylim(yr[1],yr[0])
         
-    # Add the colorbar
-    if colorlabel is None:
-        if statistic is 'count':
-            colorlabel = statistic
-        else:
-            colorlabel = statistic+'(Z)'
-    plt.colorbar(label=colorlabel)
+        # Axis titles
+        if xtitle is None:
+            xtitle = 'X'
+        plt.xlabel(xtitle)
+        if ytitle is None:
+            ytitle = 'Y'
+        plt.ylabel(ytitle)        
+        if title is None:
+            if statistic is 'count':
+                title = statistic
+            else:
+                title = statistic+'(Z)'
+        plt.title(title)
+        
+        # Add the colorbar
+        if colorlabel is None:
+            if statistic is 'count':
+                colorlabel = statistic
+            else:
+                colorlabel = statistic+'(Z)'
+        plt.colorbar(label=colorlabel)
 
-    return
+        # Save the figure
+        if save is not None:
+            plt.savefig(save,bbox_inches='tight')
+    
+    return xedges, yedges, im
 
 
-def display(im,x=None,y=None,log=False,xr=None,yr=None,noerase=False,zscale=False,zmin=None,zmax=None,
-            xtitle=None,ytitle=None,title=None,origin='lower',aspect='auto',cmap=None,figure=None):
+def display(im,x=None,y=None,log=False,xr=None,yr=None,noerase=False,zscale=False,norm=None,
+            vmin=None,vmax=None,xtitle=None,ytitle=None,title=None,origin='lower',aspect='auto',
+            xflip=False,yflip=False,cmap=None,figure=None,figsize=(8,8),save=None,colorlabel=None,
+            charsize=12,interpolation='none'):
     """
     Display an image.  The usual python convention is used where the image is [NY,NX]
 
@@ -513,9 +558,14 @@ def display(im,x=None,y=None,log=False,xr=None,yr=None,noerase=False,zscale=Fals
       Do not erase the plot window.  Default is to erase.
     zscale : boolean, optional
       Set image min/max values using the zscale algorithm.
-    zmin : float, optional
+    norm : normalize object, optional
+        The `.Normalize` instance used to scale scalar data to the [0, 1]
+        range before mapping to colors using *cmap*. By default, a linear
+        scaling mapping the lowest value to 0 and the highest to 1 is used.
+        This parameter is ignored for RGB(A) data.
+    vmin : float, optional
       Minimum image value to plot.  Default is minimum of the entire image.
-    zmax : float, optional
+    vmax : float, optional
       Maximum image value to plot.  Default is maximum of the entire image.
     xtitle : str, optional
       X-axis title  Default is no title.
@@ -527,10 +577,24 @@ def display(im,x=None,y=None,log=False,xr=None,yr=None,noerase=False,zscale=Fals
       The origin.  Default is "lower".
     aspect : str, optional
       The aspect of the plot window.  Default is "auto".
+    xflip : bool, optional
+       Flip the X-coordinate axis.  Default is False.
+    yflip : bool, optional
+       Flip the Y-coordinate axis.  Default is False.    
     cmap : str, optional
       The color map.  Default is "viridis".
     figure : int, optional
-      The figture window number.  Default is to use the current window.
+      The figure window number.  Default is to use the current window.
+    figsize : list, optional
+       Two-element list giving figure size in X and Y.  Default is (8,8).
+    save : str, optional
+       Save the figure to this file.
+    colorlabel : str, optional
+       The colorbar label.  Default is statistic+'(Z)'.
+    charsize : int, optional
+       Character size.  Default is 12.
+    interpolation : str, optional
+       The type of interpolation.  Default is 'none'.
 
     Returns
     -------
@@ -539,7 +603,7 @@ def display(im,x=None,y=None,log=False,xr=None,yr=None,noerase=False,zscale=Fals
     Example
     -------
     
-    display(im,zmin=100,zmax=5000,xr=[100,300])
+    display(im,vmin=100,vmax=5000,xr=[100,300])
 
 
     """
@@ -548,12 +612,16 @@ def display(im,x=None,y=None,log=False,xr=None,yr=None,noerase=False,zscale=Fals
     if figure is None:
         fig = plt.gcf()
     else:
-        fig = plt.figure(figure)
+        fig = plt.figure(figure,figsize=figsize)
+
+    # Character size
+    font = {'size': charsize}
+    matplotlib.rc('font', **font)
         
     if noerase is False:
         plt.clf()   # clear the plot
-
-    nx,ny = im.shape        
+        
+    ny,nx = im.shape        
         
     # No X/Y inputs
     if x is None:
@@ -587,15 +655,30 @@ def display(im,x=None,y=None,log=False,xr=None,yr=None,noerase=False,zscale=Fals
         
     # Plot the image
     #fig, ax = plt.subplots()
-    norm = None
-    if log is True:
-        norm = colors.LogNorm(vmin=zmin,vmax=zmax)
+    if log is True and norm is None:
+        norm = colors.LogNorm(vmin=vmin,vmax=vmax)
     if zscale is True:
-        zmin,zmax = zscaling(im)
+        vmin,vmax = zscaling(im)
 
+    # (left, right, bottom, top)
     extent = [xmin, xmax, ymin, ymax]
-    plt.imshow(im,cmap=cmap,norm=norm,aspect=aspect,vmin=zmin,vmax=zmax,origin=origin,extent=extent)
-        
+    if norm is not None:
+        plt.imshow(im,cmap=cmap,norm=norm,aspect=aspect,origin=origin,
+                   extent=extent,interpolation=interpolation)
+    else:
+        plt.imshow(im,cmap=cmap,aspect=aspect,vmin=vmin,vmax=vmax,origin=origin,
+                   extent=extent,interpolation=interpolation)        
+    if xflip:
+        if xr is not None:
+            plt.xlim(xr[1],xr[0])
+        else:
+            plt.xlim(xmax,xmin)            
+    if yflip:
+        if yr is not None:
+            plt.ylim(yr[1],yr[0])
+        else:
+            plt.ylim(ymax,ymin)
+            
     # Axis titles
     if xtitle is not None:
         plt.xlabel(xtitle)
@@ -605,32 +688,145 @@ def display(im,x=None,y=None,log=False,xr=None,yr=None,noerase=False,zscale=Fals
         plt.title(title)
         
     # Add the colorbar
-    plt.colorbar()
+    if colorlabel is None:
+        colorlabel = ''
+    plt.colorbar(label=colorlabel)
+
+    # Save the figure
+    if save is not None:
+        plt.savefig(save,bbox_inches='tight')
     
     return
 
 
-def plot(x,y,z=None,marker=None,log=False,noerase=False,zmin=None,zmax=None,linewidth=None,
-         xtitle=None,ytitle=None,title=None,cmap=None,alpha=None,figure=None):
-    """ Create a line or scatter plot.  like plotc.pro"""
+def plot(x,y=None,c=None,fmt=None,marker=None,fill=True,size=None,log=False,noerase=False,
+         vmin=None,vmax=None,linewidth=None,xtitle=None,ytitle=None,title=None,
+         xr=None,yr=None,cmap=None,alpha=None,figure=None,figsize=(8,8),xlog=False,ylog=False,
+         xflip=False,yflip=False,save=None,colorlabel=None,charsize=12):
+    """
+    Create a line or scatter plot.  like plotc.pro
+
+    Parameters
+    ----------
+    x : numpy array
+       Array of x-values to plot.
+    y : numpy array
+       Array of x-values to plot.
+    c : numpy array or str, optional
+       Array of values to color-code the points by, or a single color name.
+    fmt : str, optional
+      Format string for plotting.
+    marker : float/int, optional
+       Marker type.
+    fill : boolean, optional
+       Use filled markers.  Default is fill=True.
+    size : float/int, optional
+       Marker size.
+    log : boolean, optional
+       If c input, use a logarithmic scale.
+    noerase : boolean, optional
+       Do not erase the current plotting figure.  Default is to erase the figure.
+    vmin : float, optional
+       Minimum color value to plot.
+    vmax : float, optional
+       Maximum color value to plot.
+    linewidth : float, optional
+       Linewidth to use.
+    xtitle : str, optional
+       X-axis title.
+    ytitle : str, optional
+       Y-axis title.
+    title : str, optional
+       Main figure title.
+    xr : list, optional
+       X-axis minimum and maximum range.
+    yr : list, optional
+       Y-axis minimum and maximum range.
+    cmap : str, optional
+       Color map to use.
+    alpha : float, optional
+       Alpha value to use (between 0 and 1).
+    figure : int, optional
+       Figure number.  The default is to use the current figure.
+    figsize : list, optional
+       Two-element list giving figure size in X and Y.  Default is (8,8).
+    xlog : bool, optional
+       Logarithmic scaling on x-axis.  Default is linear.
+    ylog : bool, optional
+       Logarithmic scaling on y-axis.  Default is linear.
+    xflip : boolean, optional
+       Flip the x-axis.
+    yflip : boolean, optional
+       Flip the y-axis.
+    save : str, optional
+       Save the figure to this file.
+    colorlabel : str, optional
+       The colorbar label.  Default is statistic+'(Z)'.
+    charsize : int, optional
+       Character size.  Default is 12.
+
+    Returns
+    -------
+    Figure is plotted to the screen.
+
+    Example
+    -------
+
+    plotting(x,y)
+
+    """
 
     # xerr, yerr, symbol size
+
+    # No Y-input
+    if y is None:
+        y = np.array(x).copy()
+        x = np.arange(len(y))
     
     # Getting the current figure, creating a new one if necessary
     if figure is None:
         fig = plt.gcf()
     else:
-        fig = plt.figure(figure)
+        fig = plt.figure(figure,figsize=figsize)
+
+    # Character size
+    font = {'size': charsize}
+    matplotlib.rc('font', **font)
         
     if noerase is False:
         plt.clf()   # clear the plot
 
-    # Make the plot
-    norm = None
-    if log is True:
-        norm = colors.LogNorm(vmin=zmin,vmax=zmax)
-    plt.scatter(x,y,c=z,marker=marker,cmap=cmap,norm=norm,vmin=zmin,vmax=zmax,alpha=alpha,linewidth=linewidth)
+    if fill==False:
+        if c is not None:
+            edgecolor = None
+        else:
+            edgecolor = 'b'
+        facecolor = 'none'
+    else:
+        edgecolor = None
+        facecolor = None
+        
+    # Line plot
+    if (c is None or type(c) is str) and marker is None:
+        if fmt is not None:
+            plt.plot(x,y,fmt,c=c)
+        else:
+            plt.plot(x,y,c=c)
+        
+    # Scatter plot
+    else:
+        norm = None
+        if log is True:
+            norm = colors.LogNorm(vmin=vmin,vmax=vmax)
+            plt.scatter(x,y,c=c,marker=marker,s=size,cmap=cmap,norm=norm,
+                        alpha=alpha,edgecolor=edgecolor,facecolor=facecolor,
+                        linewidth=linewidth)        
+        else:
+            plt.scatter(x,y,c=c,marker=marker,s=size,cmap=cmap,norm=norm,
+                        edgecolor=edgecolor,facecolor=facecolor,vmin=vmin,
+                        vmax=vmax,alpha=alpha,linewidth=linewidth)
 
+    
     # Axis titles
     if xtitle is not None:
         plt.xlabel(xtitle)
@@ -638,13 +834,44 @@ def plot(x,y,z=None,marker=None,log=False,noerase=False,zmin=None,zmax=None,line
         plt.ylabel(ytitle)        
     if title is not None:
         plt.title(title)
-    
+
+    # Axis ranges
+    if xr is not None:
+        plt.xlim(xr)
+    if yr is not None:
+        plt.ylim(yr)
+    if xlog: plt.xscale('log')
+    if ylog: plt.yscale('log')    
+
+    # Flip axes
+    if xflip:
+        if xr is None:
+            xr = dln.minmax(x)
+            xr = [xr[0]-0.05*dln.valrange(x),xr[1]+0.05*dln.valrange(x)]            
+        plt.xlim(np.flip(xr))
+    if yflip:
+        if yr is None:
+            yr = dln.minmax(y)
+            yr = [yr[0]-0.05*dln.valrange(y),yr[1]+0.05*dln.valrange(y)]  
+        plt.ylim(np.flip(yr))
+                         
     # Add the colorbar
-    if z is not None:
-        plt.colorbar()
-    
+    if c is not None and len(c)==len(x):
+        if colorlabel is None:
+            colorlabel = ''
+        plt.colorbar(label=colorlabel)
+
+    # Save the figure
+    if save is not None:
+        plt.savefig(save,bbox_inches='tight')
+        
     return
 
+def oplot(*args,**kwargs):
+    plot(*args,**kwargs,noerase=True)
+
+def scatter(*args,marker='o',**kwargs):
+    plot(*args,marker=marker,**kwargs)
 
 
 class Cursor():
@@ -787,8 +1014,12 @@ def curpdiff(spherical=False,arcsec=False):
                 print('Delta X = %.4f %s' % (deltax,units))
                 print('Delta Y = %.4f %s' % (deltay,units))
                 print('Angle = %.4f %s' % (angle,'(CCW from Right)'))
-                slp,yoff = linear_coefficients([x1,x2],[y1,y2])
-
+                slp,yoff = dln.linear_coefficients([x1,x2],[y1,y2])
+                if yoff >= 0:
+                    print('y = %.3f*x - %.3f' % (slp,yoff))                    
+                else:
+                    print('y = %.3f*x - %.3f' % (slp,abs(yoff)))
+                          
                 curs.coords['dist'] = dist
                 curs.coords['units'] = units
                 curs.coords['deltax'] = deltax
